# -*- coding: utf-8 -*-
import logging

import click
import numpy as np
import pandas as pd
from sklearn import svm

def train(features):
    logging.info('Key ordering: %s', list(features.keys()))
    feature_values = list(features.values())
    X = np.concatenate(feature_values)
    pd.DataFrame(X).to_csv('features.csv')
    Y = np.concatenate([np.float32([i] * len(v)) for i, v in zip(range(0, len(features)), feature_values)])

    clf = svm.SVC()
    clf.fit(X, Y)
    return clf


def success_rate(classifier, features):
    from sklearn.metrics import classification_report
    from sklearn.metrics import confusion_matrix
    from sklearn.metrics import precision_score

    logging.info('Applying the classifier...')
    feature_values = list(features.values())
    X = np.concatenate(feature_values)
    Y = np.concatenate([np.float32([i] * len(v)) for i, v in zip(range(0, len(feature_values)), feature_values)])
    y_pred = classifier.predict(X)
    # logging.debug('Predictions:\n%s', list(zip(Y, y_pred)))
    logging.info('Confusion Matrix:\n%s', confusion_matrix(y_true=Y, y_pred=y_pred))
    report = classification_report(y_true=Y, y_pred=y_pred, target_names=list(features.keys()))
    logging.info('Classification Report:\n%s', report)
    return precision_score(y_true=Y, y_pred=y_pred)


@click.command()
@click.option('-d', '--dir', default='.', help='Directory of images (default: ./)')
@click.option('-g', '--load-gmm', default=False, is_flag=True, help='Load gmm dictionary from pickles')
@click.option('-n', '--number', default=5, help='Number of words in gmm dictionary')
@click.option('-l', '--limit', default=50, help='Number of images to read')
@click.option('-v', '--validation-dir', default=None, help='Directory of images (default: None)')
def main(dir, load_gmm, number, limit, validation_dir):
    """
    * Create a GMM using the training images.
    * Use this GMM to create feature vectors of training images.
    * Train an SVM on training images.
    * Predict using SVM on training images.
    """
    from aesthetics.fisher import Gmm
    from aesthetics.fisher import FisherVector
    logging.debug('dir=%s, load_gmm=%s, number=%s, limit=%s, val_dir=%s', dir, load_gmm, number, limit, validation_dir)
    gmm = Gmm(K=number)
    if load_gmm:
        gmm.load()
    else:
        gmm.generate(input_folder=dir, limit=limit)

    fisher_vector = FisherVector(gmm)
    features = fisher_vector.features(dir, limit)
    # TBD, split the features into training and validation
    classifier = train(features)
    rate = success_rate(classifier, features)
    logging.info("Self test success rate is %.2f", rate)
    if validation_dir is not None:
        validation_features = fisher_vector.features(validation_dir, limit)
        rate = success_rate(classifier, validation_features)
        logging.info("Validation test success rate is %.2f", rate)
    return 0


if __name__ == '__main__':
    import sys

    logging.basicConfig(level=logging.DEBUG)
    sys.exit(main())





import glob

import cv2  # v3.2.0
import numpy as np


class Gmm(object):
    """ K-component Gaussian Mixture Model """

    def __init__(self, K):
        """ As described in section 2.2, para 3 of reference [1] """
        self.K = K
        self.means = None
        """ Mean Vector """
        self.covariances = None
        """ Covariance Matrix """
        self.weights = None
        """ Mixture Weights """

    def generate(self, input_folder, limit):
        from aesthetics.fisher import Descriptors

        descriptor = Descriptors()
        img_descriptors = [descriptor.folder(folder, limit) for folder in sorted(glob.glob(input_folder + '/*'))]
        max_shape = np.array([v.shape[0] for v in img_descriptors]).max()
        img_descriptors = list(filter(lambda x: x is not None and x.shape[0] == max_shape, img_descriptors))
        words = np.concatenate(img_descriptors)
        print("Training GMM of size", self.K)
        self.means, self.covariances, self.weights = self.train_expectation_maximisation(words, self.K)
        # Throw away gaussians with weights that are too small:
        self.means = self.remove_too_small(self.means, self.weights)
        self.covariances = self.remove_too_small(self.covariances, self.weights)
        self.weights = self.remove_too_small(self.weights, self.weights)

        self.save()
        return self.means, self.covariances, self.weights

    def remove_too_small(self, values, weights):
        threshold = 1.0 / self.K
        return np.float32([m for k, m in zip(range(0, len(weights)), values) if weights[k] > threshold])

    def load(self, folder=''):
        import os
        files = ['means.gmm.npy', 'covariances.gmm.npy', 'weights.gmm.npy']
        self.means, self.covariances, self.weights = map(lambda file: np.load(file),
                                                         map(lambda s: os.path.join(folder, s), files))

    def save(self):
        np.save("means.gmm", self.means)
        np.save("covariances.gmm", self.covariances)
        np.save("weights.gmm", self.weights)

    @staticmethod
    def train_expectation_maximisation(descriptors, K):
        """ See reference [2] """
        em = cv2.ml.EM_create()
        em.setClustersNumber(K)
        em.trainEM(descriptors)
        return np.float32(em.getMeans()), np.float32(em.getCovs()), np.float32(em.getWeights())[0]





import click

def read_ava_data(data_folder):
    import pandas as pd
    import os

    columns = 'Index ImageId r1 r2 r3 r4 r5 r6 r7 r8 r9 r10 SemanticTag1 SemanticTag2 ChallengeId'.split()
    df = pd.read_csv(os.path.join(data_folder, 'AVA.txt'), delimiter=' ', names=columns)
    return df


def to_url(img_id):
    return 'http://www.dpchallenge.com/image.php?IMAGE_ID={img_id}'.format(img_id=img_id)


@click.command()
@click.option('--data-folder', '-d', default='./data', help='AVA Data Folder (default: ./data)')
def stats(data_folder):
    df = read_ava_data(data_folder)
    print(df.head())
    print(df.info())
    print(df.describe())


@click.command()
@click.option('--data-folder', '-d', default='./data', help='AVA Data Folder (default: ./data)')
def download_ava(data_folder):
    df = read_ava_data(data_folder)
    urls = map(to_url, df['ImageId'].values)
    for i, url in enumerate(urls):
        print(url)
        if i == 10:
            break



@click.group()
def main():
    return 0

main.add_command(download_ava, 'ava')
main.add_command(stats)




"""
Fisher Vector implementation using cv2 v3.2.0+ and python3.

References used below:
[1]: Image Classification with the Fisher Vector: https://hal.inria.fr/file/index/docid/830491/filename/journal.pdf
[2]: http://www.vlfeat.org/api/gmm-fundamentals.html
"""
import glob
import os
from concurrent.futures import ProcessPoolExecutor
from collections import OrderedDict

import numpy as np
import tqdm
from scipy.stats import multivariate_normal


class FisherVector(object):
    def __init__(self, gmm):
        self.gmm = gmm

    def features(self, folder, limit):
        folders = sorted(glob.glob(folder + "/*"))
        features = OrderedDict([(f, self.get_fisher_vectors_from_folder(f, limit)) for f in folders])
        return features

    def get_fisher_vectors_from_folder(self, folder, limit):
        files = glob.glob(folder + "/*.jpg")[:limit]

        with ProcessPoolExecutor() as pool:
            futures = pool.map(self._worker, files)
            # futures = map(self._worker, files)
            desc = 'Creating Fisher Vectors {} images of folder {}'.format(len(files), os.path.split(folder)[-1])
            futures = tqdm.tqdm(futures, total=len(files), desc=desc, unit='image')
            vectors = [f for f in futures if f is not None and len(f) > 0]
            max_shape = np.array([v.shape[0] for v in vectors]).max()
            vectors = [v for v in vectors if v.shape[0] == max_shape]
        # return np.array(vectors)    # Can't do np.float32, because all images may not have same number of features
        return np.float32(vectors)

    def _worker(self, *arg, **kwargs):
        try:
            return self.fisher_vector_of_file(*arg, **kwargs)
        except Exception as e:
            # print(e)
            # import pdb; pdb.post_mortem()
            return None

    def fisher_vector_of_file(self, filename):
        def section_fisher(img_section, full_fisher):
            sec_fisher = self.fisher_vector_of_image(img_section)
            if sec_fisher is None:
                sec_fisher = np.zeros(full_fisher.shape)
            return sec_fisher

        import cv2
        img = cv2.imread(filename)
        full_fisher = self.fisher_vector_of_image(img)
        x, _, _ = img.shape
        loc_mid = int(x / 3)
        loc_bottom = int(2 * x / 3)
        top_fisher = section_fisher(img[0:loc_mid], full_fisher)
        middle_fisher = section_fisher(img[loc_mid + 1:loc_bottom], full_fisher)
        bottom_fisher = section_fisher(img[loc_bottom + 1:x], full_fisher)
        return np.concatenate((full_fisher, top_fisher, middle_fisher, bottom_fisher))


    def fisher_vector_of_image(self, img):
        from aesthetics.fisher import Descriptors
        descriptors = Descriptors()
        img_descriptors = descriptors.image(img)
        if img_descriptors is not None:
            return self._fisher_vector(img_descriptors)
        else:
            return np.empty(0)

    def _fisher_vector(self, img_descriptors):
        """
        :param img_descriptors: X
        :return: np.array fisher vector
        """
        means, covariances, weights = self.gmm.means, self.gmm.covariances, self.gmm.weights
        s0, s1, s2 = self._likelihood_statistics(img_descriptors)
        T = img_descriptors.shape[0]
        diagonal_covariances = np.float32([np.diagonal(covariances[k]) for k in range(0, covariances.shape[0])])
        """ Refer page 4, first column of reference [1] """
        g_weights = self._fisher_vector_weights(s0, s1, s2, means, diagonal_covariances, weights, T)
        g_means = self._fisher_vector_means(s0, s1, s2, means, diagonal_covariances, weights, T)
        g_sigma = self._fisher_vector_sigma(s0, s1, s2, means, diagonal_covariances, weights, T)
        # FIXME: Weights are one dimensional here.
        fv = np.concatenate([np.concatenate(g_weights), np.concatenate(g_means), np.concatenate(g_sigma)])
        # fv = np.concatenate([g_weights, np.concatenate(means), np.concatenate(g_sigma)])
        fv = self.normalize(fv)
        return fv

    def _likelihood_statistics(self, img_descriptors):
        """
        :param img_descriptors: X
        :return: 0th order, 1st order, 2nd order statistics
                 as described by equation 20, 21, 22 in reference [1]
        """

        def likelihood_moment(x, posterior_probability, moment):
            x_moment = np.power(np.float32(x), moment) if moment > 0 else np.float32([1])
            return x_moment * posterior_probability

        def zeros(like):
            return np.zeros(like.shape).tolist()

        means, covariances, weights = self.gmm.means, self.gmm.covariances, self.gmm.weights
        normals = [multivariate_normal(mean=means[k], cov=covariances[k]) for k in range(0, len(weights))]
        """ Gaussian Normals """
        gaussian_pdfs = [np.array([g_k.pdf(sample) for g_k in normals]) for sample in img_descriptors]
        """ u(x) for equation 15, page 4 in reference 1 """
        statistics_0_order, statistics_1_order, statistics_2_order = zeros(weights), zeros(weights), zeros(weights)
        for k in range(0, len(weights)):
            for index, sample in enumerate(img_descriptors):
                posterior_probability = FisherVector.posterior_probability(gaussian_pdfs[index], weights)
                statistics_0_order[k] = statistics_0_order[k] + likelihood_moment(sample, posterior_probability[k], 0)
                statistics_1_order[k] = statistics_1_order[k] + likelihood_moment(sample, posterior_probability[k], 1)
                statistics_2_order[k] = statistics_2_order[k] + likelihood_moment(sample, posterior_probability[k], 2)

        return np.array(statistics_0_order), np.array(statistics_1_order), np.array(statistics_2_order)

    @staticmethod
    def posterior_probability(u_gaussian, weights):
        """ Implementation of equation 15, page 4 from reference [1] """
        probabilities = np.multiply(u_gaussian, weights)
        probabilities = probabilities / np.sum(probabilities)
        return probabilities

    @staticmethod
    def _fisher_vector_weights(statistics_0_order, s1, s2, means, covariances, w, T):
        """ Implementation of equation 31, page 6 from reference [1] """
        return np.float32([((statistics_0_order[k] - T * w[k]) / np.sqrt(w[k])) for k in range(0, len(w))])

    @staticmethod
    def _fisher_vector_means(s0, statistics_1_order, s2, means, sigma, w, T):
        """ Implementation of equation 32, page 6 from reference [1] """
        return np.float32([(statistics_1_order[k] - means[k] * s0[k]) /
                           (np.sqrt(w[k] * sigma[k])) for k in range(0, len(w))])

    @staticmethod
    def _fisher_vector_sigma(s0, s1, statistics_2_order, means, sigma, w, T):
        """ Implementation of equation 33, page 6 from reference [1] """
        return np.float32([(statistics_2_order[k] - 2 * means[k] * s1[k] + (means[k] * means[k] - sigma[k]) * s0[k]) /
                           (np.sqrt(2 * w[k]) * sigma[k]) for k in range(0, len(w))])

    @staticmethod
    def normalize(fisher_vector):
        """ Power normalization based on equation 30, page 5, last para; and
        is used in step 3, algorithm 1, page 6 of reference [1] """
        v = np.sign(fisher_vector) * np.sqrt(abs(fisher_vector))  # Power normalization
        return v / np.sqrt(np.dot(v, v))  # L2 Normalization





import glob
from concurrent.futures import ProcessPoolExecutor

import cv2  # v3.2.0
import numpy as np
import tqdm


class Descriptors(object):
    """ Convert image to features"""
    def __init__(self):
        self.feature_transform = None

    def folder(self, folder, limit):
        files = glob.glob(folder + "/*.jpg")[:limit]
        with ProcessPoolExecutor() as executor:
            futures = executor.map(self.image_file, files)
            futures = tqdm.tqdm(futures, total=len(files), desc='Calculating descriptors')
            descriptors = [f for f in futures]
            # descriptors = [self.image_file(file) for file in files]
        descriptors = list(filter(lambda x: x is not None, descriptors))
        return np.concatenate(descriptors)

    def image_file(self, filename):
        """ Refer section 2.2 of reference [1] """
        img = cv2.imread(filename, 0)
        return self.image(img)

    def image(self, img):
        # img = cv2.resize(img, (256, 256))
        if self.feature_transform is None:
            self.feature_transform = cv2.xfeatures2d.SIFT_create()
            # self.feature_transform = cv2.ORB_create()
        _ , descriptors = self.feature_transform.detectAndCompute(img, None)
        return descriptors


